#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef _WIN32
#define strdup _strdup
#endif

#define BUILD_GID_PLUGIN
#include "gid_plugin_import.h"
/* 
DECLARE_GID_PLUGIN_FUNCTIONS;
*/

#include "ply.h"

static char *_G_err_str = NULL;

// in ply_tcl:
void PlyProgressBar( int numerator, int divisor, const char *title = NULL);

#ifdef _DEBUG
#define CLICKS_TO_UPDATE        0x1fff
#else 
#define CLICKS_TO_UPDATE        0x1ffff
#endif

static int _G_numerator = -0;
static int _G_divisor = -12; // negative to create an indeterminate progress bar

void set_err_str( const char *str = NULL) {
  if ( _G_err_str) free( _G_err_str);
  _G_err_str = NULL;
  if ( str) {
    _G_err_str = strdup( str);
  }
}

int create_demo_triangs() {
  GiD_NewPostProcess();
  int mesh_id = GiD_NewMesh( GID_POST, GIDPOST_NEW_MESH, "Hello");
  GiD_SetColor( mesh_id, 0.5f, 0.3f, 0.8f, 0.0f);
  float lst_vertices[ 4][ 3] = { { 0.0, 0.0, 0.0}, { 1.0, 0.0, 0.0}, { 1.0, 1.0, 0.0}, { 0.0, 1.0, 0.0}};
  GiD_SetVertexPointer( mesh_id, GIDPOST_FLOAT, GIDPOST_VERTICES, 3, 4, 0, lst_vertices);
  int lst_elements[ 2][ 3] = { { 0, 1, 2}, { 2, 3, 0}};
  GiD_SetElementPointer( mesh_id, GIDPOST_FLOAT, GIDPOST_CONNECTIVITIES, GIDPOST_TRIANGLE, 2, 0, lst_elements, 0, NULL);
  
  float lst_scalar_res[ 4] = { 0.0, 1.0, 2.0, 3.0};
  int res_id = GiD_NewResult( "new analysis", 0.0, "test result", 0); // 0 --> for all meshes
  GiD_SetResultPointer( res_id, GIDPOST_FLOAT, GIDPOST_RESULTS, GIDPOST_SCALAR, GIDPOST_NODAL, 4, 0, lst_scalar_res);
  GiD_EndResult( res_id);
  GiD_EndMesh( mesh_id);
  set_err_str( "Everything fine.");
  return 0;
}

/* user's vertex and face definitions for a polygonal object */

typedef struct Vertex {
  float x,y,z;
  float nx,ny,nz;
  void *other_props;       /* other properties */
} Vertex;

typedef struct Face {
  unsigned char nverts;    /* number of vertex indices in list */
  int *verts;              /* vertex index list */
  void *other_props;       /* other properties */
} Face;

char *elem_names[] = { /* list of the kinds of elements in the user's object */
  "vertex", "face"
};

PlyProperty vert_props[] = { /* list of property information for a vertex */
  {"x", Float32, Float32, offsetof(Vertex,x), 0, 0, 0, 0},
  {"y", Float32, Float32, offsetof(Vertex,y), 0, 0, 0, 0},
  {"z", Float32, Float32, offsetof(Vertex,z), 0, 0, 0, 0},
  {"nx", Float32, Float32, offsetof(Vertex,nx), 0, 0, 0, 0},
  {"ny", Float32, Float32, offsetof(Vertex,ny), 0, 0, 0, 0},
  {"nz", Float32, Float32, offsetof(Vertex,nz), 0, 0, 0, 0},
};

PlyProperty face_props[] = { /* list of property information for a face */
  {"vertex_indices", Int32, Int32, offsetof(Face,verts),
   1, Uint8, Uint8, offsetof(Face,nverts)},
};


/*** the PLY object ***/

static PlyFile *in_ply;
static int nverts,nfaces;
static Vertex **vlist;
static Face **flist;
static PlyOtherProp *vert_other,*face_other;

// static int flip_order = 1;       /* flip order of vertices around the faces? */
// static int flip_normals = 0;     /* flip vertex normals? */

/* are normals in PLY file? */
static int has_nx = 0;
static int has_ny = 0;
static int has_nz = 0;


/******************************************************************************
Read in the PLY file from standard in.
******************************************************************************/

int read_ply_file( const char *filename) {
  int i,j;
  int elem_count;
  char *elem_name;
  
  /*** Read in the original PLY object ***/
  FILE *f_in = fopen( filename, "rb");
  if ( !f_in)
    return 0;
  in_ply = read_ply( f_in); // close_ply already fclose's f_in

  if ( !in_ply) {
    set_err_str( "Problems reading the file");
    fclose( f_in);
    return 0;
  }
  
  /* examine each element type that is in the file (vertex, face) */

  for (i = 0; i < in_ply->num_elem_types; i++) {
        
    /* prepare to read the i'th list of elements */
    elem_name = setup_element_read_ply (in_ply, i, &elem_count);
    if ( elem_name == NULL) {
      set_err_str( "Problems reading the file");
      close_ply (in_ply);
      free_ply( in_ply);
      in_ply = NULL;
      return 0;
    }
    char buf[ 1024];
    sprintf( buf, "PLY: Reading %s ...", elem_name);
    PlyProgressBar( _G_numerator++, _G_divisor, buf);
    
    if (equal_strings ("vertex", elem_name)) {

      /* create a vertex list to hold all the vertices */
      vlist = (Vertex **) malloc (sizeof (Vertex *) * elem_count);
      nverts = elem_count;

      /* set up for getting vertex elements */
      /* (we want x,y,z) */

      setup_property_ply (in_ply, &vert_props[0]);
      setup_property_ply (in_ply, &vert_props[1]);
      setup_property_ply (in_ply, &vert_props[2]);

      /* we also want normal information if it is there (nx,ny,nz) */

      for (j = 0; j < in_ply->elems[i]->nprops; j++) {
	PlyProperty *prop;
	prop = in_ply->elems[i]->props[j];
	if (equal_strings ("nx", prop->name)) {
	  setup_property_ply (in_ply, &vert_props[3]);
	  has_nx = 1;
	}
	if (equal_strings ("ny", prop->name)) {
	  setup_property_ply (in_ply, &vert_props[4]);
	  has_ny = 1;
	}
	if (equal_strings ("nz", prop->name)) {
	  setup_property_ply (in_ply, &vert_props[5]);
	  has_nz = 1;
	}
      }

      /* also grab anything else that we don't need to know about */

      vert_other = get_other_properties_ply (in_ply, 
					     offsetof(Vertex,other_props));

      /* grab the vertex elements and store them in our list */

      for (j = 0; j < elem_count; j++) {
        vlist[j] = (Vertex *) malloc (sizeof (Vertex));
        if ( get_element_ply (in_ply, (void *) vlist[j]) == -1) {
          set_err_str( "Problems reading vertices of the file");
          close_ply (in_ply);
          free_ply( in_ply);
          in_ply = NULL;
          return 0;
        }

        if ( !( j & CLICKS_TO_UPDATE)) {
          PlyProgressBar( _G_numerator++, _G_divisor);
        }
      }
    }
    else if (equal_strings ("face", elem_name)) {

      /* create a list to hold all the face elements */
      flist = (Face **) malloc (sizeof (Face *) * elem_count);
      nfaces = elem_count;

      /* set up for getting face elements */
      /* (all we need are vertex indices) */

      setup_property_ply (in_ply, &face_props[0]);
      face_other = get_other_properties_ply (in_ply, 
					     offsetof(Face,other_props));

      /* grab all the face elements and place them in our list */

      for (j = 0; j < elem_count; j++) {
        flist[j] = (Face *) malloc (sizeof (Face));
        if ( get_element_ply (in_ply, (void *) flist[j]) == -1) {
          set_err_str( "Problems reading faces of the file");
          close_ply (in_ply);
          free_ply( in_ply);
          in_ply = NULL;
          return 0;
        }

        if ( !( j & CLICKS_TO_UPDATE)) {
          PlyProgressBar( _G_numerator++, _G_divisor);
        }
      }
    }
    else  /* all non-vertex and non-face elements are grabbed here */
      get_other_element_ply (in_ply);
  }

  /* close the file */
  /* (we won't free up the memory for in_ply because we will use it */
  /*  to help describe the file that we will write out) */

  close_ply (in_ply);
  free_ply( in_ply);
  in_ply = NULL;
  return nverts;
}

template< int N>
int *get_connectivity( int num_elements, Face **lst_faces, int n_faces) {
  int *connectivity = ( int *)malloc( num_elements * N * sizeof( int));
  int idx_connec = 0;
  for ( int i = 0 ; i < n_faces; i++) {
    if ( lst_faces[ i]->nverts == N) {
      for ( int j = 0; j < N; j++) {
        connectivity[ idx_connec++] = lst_faces[ i]->verts[ j];
      }
    }

    if ( !( i & CLICKS_TO_UPDATE)) {
      PlyProgressBar( _G_numerator++, _G_divisor);
    }
  }
  return connectivity;
}

// #define StartType  0
// #define Int8       1
// #define Int16      2
// #define Int32      3
// #define Uint8      4
// #define Uint16     5
// #define Uint32     6
// #define Float32    7
// #define Float64    8
// #define EndType    9
static float plyType2Float( unsigned char *data, int ply_type) {
  float ret = 0.0;
  switch( ply_type) {
    case Int8:
      ret = ( float)*( ( char *)data);
      break;
    case Int16:
      ret = ( float)*( ( short int *)data);
      break;
    case Int32:
      ret = ( float)*( ( int *)data);
      break;
    case Uint8:
      ret = ( float)*( ( unsigned char *)data);
      break;
    case Uint16:
      ret = ( float)*( ( unsigned short int *)data);
      break;
    case Uint32:
      ret = ( float)*( ( unsigned int *)data);
      break;
    case Float32:
      ret = *( ( float *)data);
      break;
    case Float64:
      ret = ( float)*( ( double *)data);
      break;
  }
  return ret;
}

int ply_2_gid( const char *filename) {
  const char *name = strrchr( filename, '/');
#ifdef _WIN32
  if ( !name) name = strrchr( filename, '\\');
#endif // _WIN32
  if ( !name) name = filename;
  else name++;

  int ret = read_ply_file( filename);

  int err = 0;
  if ( ret && nverts && nfaces) {
    err = 0;
    PlyProgressBar( _G_numerator++, _G_divisor, "PLY: passing information to GiD ...");
    GiD_NewPostProcess();
    int mesh_id = GiD_NewMesh( GID_POST, GIDPOST_NEW_MESH, name);
    GiD_SetColor( mesh_id, 0.8f, 0.8f, 0.8f, 1.0f);
    float *lst_vertices = ( float *)malloc( 3 * nverts * sizeof( float));
    int idx_v = 0;
    for ( int i = 0; i < nverts; i++) {
      lst_vertices[ idx_v++] = vlist[ i]->x;
      lst_vertices[ idx_v++] = vlist[ i]->y;
      lst_vertices[ idx_v++] = vlist[ i]->z;
    }
    GiD_SetVertexPointer( mesh_id, GIDPOST_FLOAT, GIDPOST_VERTICES, 3, nverts, 0, lst_vertices);
    free( lst_vertices);
    int num_elements[ 11] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    for ( int i = 0; i < nfaces; i++) {
      int num_verts = flist[ i]->nverts;
      if ( num_verts > 10) num_verts = 10;
      num_elements[ num_verts]++;
    }
    // create lines:
    if ( num_elements[ 2]) {
      int *lines = get_connectivity< 2>( num_elements[ 2], flist, nfaces);
      GiD_SetElementPointer( mesh_id, GIDPOST_FLOAT, GIDPOST_CONNECTIVITIES, GIDPOST_LINE,
                             num_elements[ 2], 0, lines, 0, NULL);
      free( lines);
    }
    // create triangles:
    if ( num_elements[ 3]) {
      int *triangles = get_connectivity< 3>( num_elements[ 3], flist, nfaces);
      GiD_SetElementPointer( mesh_id, GIDPOST_FLOAT, GIDPOST_CONNECTIVITIES, GIDPOST_TRIANGLE,
                             num_elements[ 3], 0, triangles, 0, NULL);
      free( triangles);
    }
    // create quadrilaterals:
    if ( num_elements[ 4]) {
      int *quads = get_connectivity< 4>( num_elements[ 4], flist, nfaces);
      GiD_SetElementPointer( mesh_id, GIDPOST_FLOAT, GIDPOST_CONNECTIVITIES, GIDPOST_QUADRILATERAL,
                             num_elements[ 4], 0, quads, 0, NULL);
      free( quads);
    }

    GiD_EndMesh( mesh_id);
    
    // now the other properties:
    if ( has_nx && has_ny && has_nz) {
      float *lst_normals = ( float *)malloc( 3 * nverts * sizeof( float));
      int idx_v = 0;
      for ( int i = 0; i < nverts; i++) {
        lst_normals[ idx_v++] = vlist[ i]->nx;
        lst_normals[ idx_v++] = vlist[ i]->ny;
        lst_normals[ idx_v++] = vlist[ i]->nz;
      }
      int res_id = GiD_NewResult( "Geometry", 0.0, "Normals", 0);
      GiD_SetResultPointer( res_id, GIDPOST_FLOAT, GIDPOST_RESULTS, GIDPOST_VECTOR_3, GIDPOST_NODAL, nverts, 0, lst_normals);
      free( lst_normals);
      GiD_EndResult( res_id);
    }

    bool both_properties = vert_other && vert_other->nprops && face_other && face_other->nprops;
    if ( vert_other && vert_other->nprops) {
      float *lst_prop = ( float *)malloc( nverts * sizeof( float));
      size_t offset_ply_types[ EndType] = { 0, 1, 2, 4, 1, 2, 4, 4, 8};
      size_t offset_prop = 0;
      char tmp_name[ 8192];
      for ( int iprop = 0; iprop < vert_other->nprops; iprop++) {
        PlyProperty *cur_prop = vert_other->props[ iprop];
        if( both_properties) {
          sprintf( tmp_name, "vertex//%s", cur_prop->name);
        } else {
          strcpy( tmp_name, cur_prop->name);
        }
        int res_id = GiD_NewResult( "Geometry", 0.0, tmp_name, 0);
        for ( int i = 0; i < nverts; i++) {
          unsigned char *ptr_cur_prop = ( unsigned char *)( vlist[ i]->other_props) + cur_prop->offset; // offset_prop;
          lst_prop[ i] = plyType2Float( ptr_cur_prop, cur_prop->internal_type);
        }
        GiD_SetResultPointer( res_id, GIDPOST_FLOAT, GIDPOST_RESULTS, GIDPOST_SCALAR, GIDPOST_NODAL, nverts, 0, lst_prop);
        GiD_EndResult( res_id);
        offset_prop += offset_ply_types[ cur_prop->internal_type];
      }
      free( lst_prop);
    }
    if ( face_other && face_other->nprops) {
      float *lst_prop = ( float *)malloc( nfaces * sizeof( float));
      size_t offset_ply_types[ EndType] = { 0, 1, 2, 4, 1, 2, 4, 4, 8};
      size_t offset_prop = 0;
      char tmp_name[ 8192];
      for ( int iprop = 0; iprop < face_other->nprops; iprop++) {
        PlyProperty *cur_prop = face_other->props[ iprop];
        if( both_properties) {
          sprintf( tmp_name, "face//%s", cur_prop->name);
        } else {
          strcpy( tmp_name, cur_prop->name);
        }
        int res_id = GiD_NewResult( "Geometry", 0.0, tmp_name, 0);
        for ( int i = 0; i < nfaces; i++) {
          unsigned char *ptr_cur_prop = ( unsigned char *)( flist[ i]->other_props) + cur_prop->offset; // offset_prop;
          lst_prop[ i] = plyType2Float( ptr_cur_prop, cur_prop->internal_type);
	  // if ( ptr_cur_prop) free( ptr_cur_prop); // free local other_prop
        }
        GiD_SetResultPointer( res_id, GIDPOST_FLOAT, GIDPOST_RESULTS, GIDPOST_SCALAR, GIDPOST_ELEMENTAL, nfaces, 0, lst_prop);
        GiD_EndResult( res_id);
        offset_prop += offset_ply_types[ cur_prop->internal_type];
	// if ( cur_prop) free( cur_prop); // free other_prop
      }
      free( lst_prop);
    }

    PlyProgressBar( _G_numerator++, _G_divisor, "PLY: freeing memory ...");
    // liberamos memoria:
    
    for ( int i = 0; i < nverts; i++) {
      if ( vert_other && vert_other->nprops) {
        free( vlist[ i]->other_props);
      }
      free( vlist[ i]);

      if ( !( i & CLICKS_TO_UPDATE)) {
        PlyProgressBar( _G_numerator++, _G_divisor);
      }
    }
    free( vlist);
    
    PlyProgressBar( _G_numerator++, _G_divisor);

    if ( vert_other) {
      for ( int ip = 0; ip < vert_other->nprops; ip++) {
	PlyProperty *cur_prop = vert_other->props[ ip];
	if ( cur_prop) {
	  if ( cur_prop->name) {
	    free( cur_prop->name);
	  }
	  free( cur_prop);
	}
      }
      if ( vert_other->name) {
	free( vert_other->name);
	vert_other->name = NULL;
      }
      free( vert_other->props);
      vert_other->props = NULL;
      free( vert_other);
      vert_other = NULL;
    }

    PlyProgressBar( _G_numerator++, _G_divisor);

    for ( int i = 0; i < nfaces; i++) {
      free( flist[ i]->verts); // they are internally allocated inside ply.c
      if ( face_other && face_other->nprops) {
        free( flist[ i]->other_props);
      }
      free( flist[ i]);

      if ( !( i & CLICKS_TO_UPDATE)) {
        PlyProgressBar( _G_numerator++, _G_divisor);
      }
    }
    free( flist);
    
    PlyProgressBar( _G_numerator++, _G_divisor);

    if ( face_other) {
      for ( int ip = 0; ip < face_other->nprops; ip++) {
	PlyProperty *cur_prop = face_other->props[ ip];
	if ( cur_prop) {
	  if ( cur_prop->name) {
	    free( cur_prop->name);
	  }
	  free( cur_prop);
	}
      }
      if ( face_other->name) {
	free( face_other->name);
	face_other->name = NULL;
      }
      free( face_other->props);
      face_other->props = NULL;
      free( face_other);
      face_other = NULL;
    }

    nverts = 0;
    nfaces = 0;
    
  } else {
    err = 1;
  }
  
  
  // create_demo_triangs();

  return err;
}

/*************************************************************************
 * Interface functions to GiD
 *************************************************************************/

// must be of type:
// typedef int ( *_t_fncGiD_PluginImport)( const char *filename);
// called by GiD
extern "C" GID_DLL_EXPORT int GiD_PostImportFile( const char *filename) {

  char *path2file = strdup( filename);
  char *s = strrchr( path2file, '/');
  if ( !s) s = strrchr( path2file, '\\');
  if ( s) *( ++s) = '\0';
  else strcpy( path2file, "./");
  
  bool res = true;
  int ret_val = 1; // 1 - error, 0 - success

  // extern "C" GID_DLL_EXPORT void hello() {
  // printf( "-->hello.\n");
  // printf( "-->hello: %d\n", hola);
  printf( "importing %s: %s\n", filename, res ? "Ok" : "can not be read");
  _G_numerator = 0;
  PlyProgressBar( _G_numerator++, _G_divisor, "Importing PLY...");
  if ( res) {
    // ret_val = create_demo_triangs();
    ret_val = ply_2_gid( filename);
  } else {
    char buf[ 8192];
    sprintf( buf, "Problems reading the Polygon '%s'.", filename);
    set_err_str( buf);
    ret_val = 1;
  }
  free( path2file);
  PlyProgressBar( 0, 0, "PLY: Done");
  return ret_val;
}

extern "C" GID_DLL_EXPORT const char *GiD_PostImportGetLibName( void) {
  return "Polygon File Format Import (tcl)"; 
}

extern "C" GID_DLL_EXPORT const char *GiD_PostImportGetFileExtensions( void) {
  return "{{PLY files} {.ply}} {{All files} {*}}";
}

extern "C" GID_DLL_EXPORT const char *GiD_PostImportGetDescription( void) {
  return "PLY-tcl ( Polygon File Format) import plugin for GiD with progressbar"; 
}

extern "C" GID_DLL_EXPORT const char *GiD_PostImportGetErrorStr( void) {
  return _G_err_str; 
}
